{% extends "basic.j2" %}

{% block head %}
<link href="/static/css/c3.min.css" rel="stylesheet" type="text/css">
{% endblock head %}

{% block body %}
<script src="/static/js/d3.min.js" charset="utf-8"></script>
<script src="/static/js/c3.min.js"></script>
<script src="/static/js/gnocchi.js"></script>
<div class="container-fluid">
  <div class="page-header">
    <h1>Measures</h1>
  </div>

  <div id="chart"></div>

</div>

<script>
$(function () {
    drawGraph('#chart', {{ data|tojson }})
});
</script>
{% endblock body %}
